import { BotGenerator } from "@spt-aki/generators/BotGenerator";
import { BotDifficultyHelper } from "@spt-aki/helpers/BotDifficultyHelper";
import { BotHelper } from "@spt-aki/helpers/BotHelper";
import { ProfileHelper } from "@spt-aki/helpers/ProfileHelper";
import { WeightedRandomHelper } from "@spt-aki/helpers/WeightedRandomHelper";
import { IBotBase } from "@spt-aki/models/eft/common/tables/IBotBase";
import { IBotType } from "@spt-aki/models/eft/common/tables/IBotType";
import { BotGenerationDetails } from "@spt-aki/models/spt/bots/BotGenerationDetails";
import { ILogger } from "@spt-aki/models/spt/utils/ILogger";
import { ConfigServer } from "@spt-aki/servers/ConfigServer";
import { DatabaseServer } from "@spt-aki/servers/DatabaseServer";
import { BotEquipmentFilterService } from "@spt-aki/services/BotEquipmentFilterService";
import { LocalisationService } from "@spt-aki/services/LocalisationService";
import { SeasonalEventService } from "@spt-aki/services/SeasonalEventService";
import { HashUtil } from "@spt-aki/utils/HashUtil";
import { JsonUtil } from "@spt-aki/utils/JsonUtil";
import { RandomUtil } from "@spt-aki/utils/RandomUtil";
import { TimeUtil } from "@spt-aki/utils/TimeUtil";
import { BotInventoryGenerator } from "@spt-aki/generators/BotInventoryGenerator";
import { BotLevelGenerator } from "@spt-aki/generators/BotLevelGenerator";
import { inject, injectable} from "tsyringe";


@injectable()
export class customBotGen extends BotGenerator
{
	constructor(
		@inject("WinstonLogger") protected logger: ILogger,
        @inject("HashUtil") protected hashUtil: HashUtil,
        @inject("RandomUtil") protected randomUtil: RandomUtil,
        @inject("TimeUtil") protected timeUtil: TimeUtil,
        @inject("JsonUtil") protected jsonUtil: JsonUtil,
        @inject("ProfileHelper") protected profileHelper: ProfileHelper,
        @inject("DatabaseServer") protected databaseServer: DatabaseServer,
        @inject("BotInventoryGenerator") protected botInventoryGenerator: BotInventoryGenerator,
        @inject("BotLevelGenerator") protected botLevelGenerator: BotLevelGenerator,
        @inject("BotEquipmentFilterService") protected botEquipmentFilterService: BotEquipmentFilterService,
        @inject("WeightedRandomHelper") protected weightedRandomHelper: WeightedRandomHelper,
        @inject("BotHelper") protected botHelper: BotHelper,
        @inject("BotDifficultyHelper") protected botDifficultyHelper: BotDifficultyHelper,
        @inject("SeasonalEventService") protected seasonalEventService: SeasonalEventService,
        @inject("LocalisationService") protected localisationService: LocalisationService,
        @inject("ConfigServer") protected configServer: ConfigServer
	)
	{
		super(logger, hashUtil, randomUtil, timeUtil, jsonUtil, profileHelper, databaseServer, botInventoryGenerator, botLevelGenerator, botEquipmentFilterService, weightedRandomHelper, botHelper, botDifficultyHelper, seasonalEventService, localisationService, configServer);
	}
	protected override generate(sessionId: string, bot: IBotBase, botJsonTemplate: IBotType, botGenerationDetails: BotGenerationDetails): IBotBase
	{
		const botRole = botGenerationDetails.role.toLowerCase();
        const botLevel = this.botLevelGenerator.generateBotLevel(botJsonTemplate.experience.level, botGenerationDetails, bot);
        var botPMC = false;
        for (let type in botGenerationDetails.conditions)
		{
			let roles = botGenerationDetails.conditions[type]
			if(roles.Role == "pmcBot"){
                botPMC = true;
            }
		}

        bot.Info.Nickname = this.generateBotNickname(botJsonTemplate, botGenerationDetails.isPlayerScav, botRole, sessionId);
        if (!botGenerationDetails.isPlayerScav)
        {
            this.botEquipmentFilterService.getBotEquipmentBlacklist(botRole, botLevel.level);
            this.botEquipmentFilterService.filterBotEquipment(botJsonTemplate, botLevel.level, botPMC, botRole);
        }

        const skipChristmasItems = !this.seasonalEventService.christmasEventEnabled();
        if (skipChristmasItems)
        {
            this.seasonalEventService.removeChristmasItemsFromBotInventory(botJsonTemplate.inventory, botGenerationDetails.role);
        }

        // Remove hideout data if bot is not a PMC or pscav
        if (!(botGenerationDetails.isPmc || botGenerationDetails.isPlayerScav))
        {
            bot.Hideout = null;
        }

        bot.Info.Experience = botLevel.exp;
        bot.Info.Level = botLevel.level;
        bot.Info.Settings.Experience = this.randomUtil.getInt(botJsonTemplate.experience.reward.min, botJsonTemplate.experience.reward.max);
        bot.Info.Settings.StandingForKill = botJsonTemplate.experience.standingForKill;
        bot.Info.Voice = this.randomUtil.getArrayValue(botJsonTemplate.appearance.voice);
        bot.Health = this.generateHealth(botJsonTemplate.health, bot.Info.Side === "Savage");
        bot.Skills = this.generateSkills(<any>botJsonTemplate.skills); // TODO: fix bad type, bot jsons store skills in dict, output needs to be array
        

        // START CUSTOM BOT GEN CODE
        const dbServer = this.databaseServer.getTables();
        const addClothingToBots = dbServer.bots.botClothingGen.addClothingToBots;
        const clothingTags = dbServer.bots.botClothingGen.tags;
        const clothingTagWeights = dbServer.bots.botClothingGen.weights;
        const customClothingBotTypes = dbServer.bots.botClothingGen.botTypes;
        const customClothingChance = dbServer.bots.botClothingGen.rollWeights;

        function defaultClothingGen(weightedRandomHelper) {
            bot.Customization.Body = weightedRandomHelper.getWeightedInventoryItem(botJsonTemplate.appearance.body);
            bot.Customization.Feet = weightedRandomHelper.getWeightedInventoryItem(botJsonTemplate.appearance.feet);
        }

        // tags without atleast 1 item in bottoms AND top will be rerolled.
        // reroll is our base case and stops the function from endlessly calling itself if all the tags dont have a matching set.
        function getMatchingSet(weightedRandomHelper, reroll = 0) {
            const chosenTag = weightedRandomHelper.getWeightedInventoryItem(clothingTagWeights);
            if ("tops" in clothingTags[chosenTag] && "bottoms" in clothingTags[chosenTag]){
                bot.Customization.Body = weightedRandomHelper.getWeightedInventoryItem(clothingTags[chosenTag]["tops"]);
                bot.Customization.Feet = weightedRandomHelper.getWeightedInventoryItem(clothingTags[chosenTag]["bottoms"]);
            } else {
                if (Object.keys(clothingTags).length >= 2 && reroll <= 3) {
                    reroll++;
                    getMatchingSet(weightedRandomHelper, reroll);
                } else {
                    defaultClothingGen(weightedRandomHelper);
                }
            }
        }

        if (addClothingToBots) {
            const rolledWeightValue = this.weightedRandomHelper.getWeightedInventoryItem(customClothingChance);
            if (rolledWeightValue == "custom" && customClothingBotTypes.includes(botRole) && Object.keys(clothingTagWeights).length >= 1)
            {
                getMatchingSet(this.weightedRandomHelper);
            } else {
                defaultClothingGen(this.weightedRandomHelper);
            }
        } else {
            defaultClothingGen(this.weightedRandomHelper);
        }
        // END CUSTOM BOT GEN CODE

        bot.Customization.Head = this.randomUtil.getArrayValue(botJsonTemplate.appearance.head);
        bot.Customization.Hands = this.randomUtil.getArrayValue(botJsonTemplate.appearance.hands);
        bot.Inventory = this.botInventoryGenerator.generateInventory(sessionId, botJsonTemplate, botRole, botGenerationDetails.isPmc, botLevel.level);

        if (this.botHelper.isBotPmc(botRole))
        {
            this.getRandomisedGameVersionAndCategory(bot.Info);
            bot = this.generateDogtag(bot);
        }



        // generate new bot ID
        bot = this.generateId(bot);

        // generate new inventory ID
        bot = this.generateInventoryID(bot);

        return bot;
	}

}
